import abc
import enum
import io
import json
import typing as ta

from ... import lang
from . import consts


I = ta.TypeVar('I')
Scalar: ta.TypeAlias = bool | int | float | str | None

SCALAR_TYPES: tuple[type, ...] = (bool, int, float, str, type(None))

MULTILINE_SEPARATORS = consts.Separators(',', ': ')


##


class AbstractJsonRenderer(lang.Abstract, ta.Generic[I]):
    class State(enum.Enum):
        VALUE = enum.auto()
        KEY = enum.auto()

    def __init__(
            self,
            indent: int | str | None = None,
            separators: tuple[str, str] | None = None,
            sort_keys: bool = False,
            style: ta.Callable[[ta.Any, State], tuple[str, str]] | None = None,
            ensure_ascii: bool = True,
    ) -> None:
        super().__init__()

        self._sort_keys = sort_keys
        self._style = style
        self._ensure_ascii = ensure_ascii

        if isinstance(indent, (str, int)):
            self._indent = (' ' * indent) if isinstance(indent, int) else indent
            self._endl = '\n'
            if separators is None:
                separators = MULTILINE_SEPARATORS
        elif indent is None:
            self._indent = self._endl = ''
            if separators is None:
                separators = consts.PRETTY_SEPARATORS
        else:
            raise TypeError(indent)
        self._comma, self._colon = separators

        self._level = 0
        self._indent_cache: dict[int, str] = {}

    _literals: ta.ClassVar[ta.Mapping[ta.Any, str]] = {
        True: 'true',
        False: 'false',
        None: 'null',
    }

    def _get_indent(self) -> str:
        if not self._indent:
            return ''

        if not self._level:
            return self._endl

        try:
            return self._indent_cache[self._level]
        except KeyError:
            pass

        ret = self._endl + (self._indent * self._level)
        self._indent_cache[self._level] = ret
        return ret

    def _format_scalar(self, o: Scalar) -> str:
        if o is None or isinstance(o, bool):
            return self._literals[o]

        elif isinstance(o, (str, int, float)):
            return json.dumps(o, ensure_ascii=self._ensure_ascii)

        else:
            raise TypeError(o)

    @classmethod
    @abc.abstractmethod
    def render_str(cls, i: I, /, **kwargs: ta.Any) -> str:
        raise NotImplementedError


##


class JsonRendererOut(ta.Protocol):
    def write(self, s: str) -> ta.Any: ...


class JsonRenderer(AbstractJsonRenderer[ta.Any]):
    def __init__(
            self,
            out: JsonRendererOut,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__(**kwargs)

        self._out = out

    def _write(self, s: str) -> None:
        if s:
            self._out.write(s)

    def _write_indent(self) -> None:
        self._write(self._get_indent())

    def _render(
            self,
            o: ta.Any,
            state: AbstractJsonRenderer.State = AbstractJsonRenderer.State.VALUE,
    ) -> None:
        if self._style is not None:
            pre, post = self._style(o, state)
            self._write(pre)
        else:
            post = None

        if isinstance(o, SCALAR_TYPES):
            self._write(self._format_scalar(o))  # type: ignore

        elif isinstance(o, ta.Mapping):
            self._write('{')
            self._level += 1
            items = list(o.items())
            if self._sort_keys:
                items.sort(key=lambda t: t[0])
            for i, (k, v) in enumerate(items):
                if i:
                    self._write(self._comma)
                self._write_indent()
                self._render(k, AbstractJsonRenderer.State.KEY)
                self._write(self._colon)
                self._render(v)
            self._level -= 1
            if o:
                self._write_indent()
            self._write('}')

        elif isinstance(o, ta.Sequence):
            self._write('[')
            self._level += 1
            for i, e in enumerate(o):
                if i:
                    self._write(self._comma)
                self._write_indent()
                self._render(e)
            self._level -= 1
            if o:
                self._write_indent()
            self._write(']')

        else:
            raise TypeError(o)

        if post:
            self._write(post)

    def render(self, o: ta.Any) -> None:
        self._render(o)

    @classmethod
    def render_str(cls, i: ta.Any, /, **kwargs: ta.Any) -> str:
        out = io.StringIO()
        cls(out, **kwargs).render(i)
        return out.getvalue()



import collections


class ScopedChainDict:
    """
    Dictionary that can delegate lookups to multiple dicts. This provides a basic get/set dict interface that is backed
    by multiple dicts.  Each dict is searched from the top most (most recently pushed) scope dict until a match is
    found.
    """

    def __init__(self, *scopes):
        super().__init__()

        # The scopes are evaluated starting at the top of the stack (the most recently pushed scope via .push_scope()).
        # If we use a normal list() and push/pop scopes by adding/removing to the end of the list, we'd have to always
        # call reversed(self._scopes) whenever we resolve a key, because the end of the list is the top of the stack.
        # To avoid this, we're using a deque so we can append to the front of the list via .appendleft() in constant
        # time, and iterate over scopes without having to do so with a reversed() call each time.
        self._scopes = collections.deque(scopes)

    def __getitem__(self, key):
        for scope in self._scopes:
            if key in scope:
                return scope[key]
        raise KeyError(key)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def push_scope(self, scope):
        self._scopes.appendleft(scope)

    def pop_scope(self):
        self._scopes.popleft()



from ... import dataclasses as dc
from ... import lang


@dc.dataclass(frozen=True)
class KnownError:
    code: int
    message: str
    meaning: str


class KnownErrors(lang.Namespace):
    PARSE_ERROR = KnownError(-32700, 'Parse error', 'Invalid JSON was received by the server.')
    INVALID_REQUEST = KnownError(-32600, 'Invalid Request', 'The JSON sent is not a valid Request object.')
    METHOD_NOT_FOUND = KnownError(-32601, 'Method not found', 'The method does not exist / is not available.')
    INVALID_PARAMS = KnownError(-32602, 'Invalid params', 'Invalid method parameter(s).')
    INTERNAL_ERROR = KnownError(-32603, 'Internal error', 'Internal JSON-RPC error.')


CUSTOM_ERROR_BASE = -32000



import typing as ta


class Node(ta.TypedDict):
    type: str
    children: list['Node']
    value: ta.NotRequired[ta.Any]


def arithmetic_unary(operator, expression) -> Node:
    return {'type': 'arithmetic_unary', 'children': [expression], 'value': operator}


def arithmetic(operator, left, right) -> Node:
    return {'type': 'arithmetic', 'children': [left, right], 'value': operator}


def assign(name, expr) -> Node:
    return {'type': 'assign', 'children': [expr], 'value': name}


def comparator(name, first, second) -> Node:
    return {'type': 'comparator', 'children': [first, second], 'value': name}


def current_node() -> Node:
    return {'type': 'current', 'children': []}


def root_node() -> Node:
    return {'type': 'root', 'children': []}


def expref(expression) -> Node:
    return {'type': 'expref', 'children': [expression]}


def function_expression(name, args) -> Node:
    return {'type': 'function_expression', 'children': args, 'value': name}


def field(name) -> Node:
    return {'type': 'field', 'children': [], 'value': name}


def filter_projection(left, right, comparator) -> Node:
    return {'type': 'filter_projection', 'children': [left, right, comparator]}


def flatten(node) -> Node:
    return {'type': 'flatten', 'children': [node]}


def identity() -> Node:
    return {'type': 'identity', 'children': []}


def index(index) -> Node:
    return {'type': 'index', 'value': index, 'children': []}


def index_expression(children) -> Node:
    return {'type': 'index_expression', 'children': children}


def key_val_pair(key_name, node) -> Node:
    return {'type': 'key_val_pair', 'children': [node], 'value': key_name}


def let_expression(bindings, expr) -> Node:
    return {'type': 'let_expression', 'children': [*bindings, expr]}


def literal(literal_value) -> Node:
    return {'type': 'literal', 'value': literal_value, 'children': []}


def multi_select_dict(nodes) -> Node:
    return {'type': 'multi_select_dict', 'children': nodes}


def multi_select_list(nodes) -> Node:
    return {'type': 'multi_select_list', 'children': nodes}


def or_expression(left, right) -> Node:
    return {'type': 'or_expression', 'children': [left, right]}


def and_expression(left, right) -> Node:
    return {'type': 'and_expression', 'children': [left, right]}


def not_expression(expr) -> Node:
    return {'type': 'not_expression', 'children': [expr]}


def pipe(left: Node, right: Node) -> Node:
    return {'type': 'pipe', 'children': [left, right]}


def projection(left: Node, right: Node) -> Node:
    return {'type': 'projection', 'children': [left, right]}


def subexpression(children) -> Node:
    return {'type': 'subexpression', 'children': children}


def slice(start, end, step) -> Node:  # noqa
    return {'type': 'slice', 'children': [start, end, step]}


def value_projection(left: Node, right: Node) -> Node:
    return {'type': 'value_projection', 'children': [left, right]}


def variable_ref(name: str) -> Node:
    return {'type': 'variable_ref', 'children': [], 'value': name}



import asyncio
import contextlib
import functools
import typing as ta


CallableT = ta.TypeVar('CallableT', bound=ta.Callable)


def asyncio_once(fn: CallableT) -> CallableT:
    future = None

    @functools.wraps(fn)
    async def inner(*args, **kwargs):
        nonlocal future
        if not future:
            future = asyncio.create_task(fn(*args, **kwargs))
        return await future

    return ta.cast(CallableT, inner)


def get_real_current_loop() -> asyncio.AbstractEventLoop | None:
    return asyncio.get_event_loop_policy()._local._loop  # type: ignore  # noqa


def drain_tasks(loop=None):
    if loop is None:
        loop = get_real_current_loop()

    while loop._ready or loop._scheduled:  # noqa
        loop._run_once()  # noqa


@contextlib.contextmanager
def draining_asyncio_tasks() -> ta.Iterator[None]:
    loop = get_real_current_loop()
    try:
        yield
    finally:
        if loop is not None:
            drain_tasks(loop)  # noqa



from .. import reflect as rfl


class MarshalError(Exception):
    pass


class UnhandledTypeError(MarshalError):
    @property
    def rty(self) -> rfl.Type:
        return self.args[0]


class ForbiddenTypeError(MarshalError):
    @property
    def rty(self) -> rfl.Type:
        return self.args[0]



import dataclasses as dc
import json
import typing as ta

from .... import lang
from .... import term
from ..render import JsonRenderer
from ..stream.parse import JsonStreamParserEvent
from ..stream.render import StreamJsonRenderer


##


@dc.dataclass(frozen=True, kw_only=True)
class RenderingOptions:
    indent: int | str | None = None
    separators: tuple[str, str] | None = None
    sort_keys: bool = False
    raw: bool = False
    unicode: bool = False
    color: bool = False


def make_render_kwargs(opts: RenderingOptions) -> ta.Mapping[str, ta.Any]:
    return dict(
        indent=opts.indent,
        separators=opts.separators,
        sort_keys=opts.sort_keys,
        ensure_ascii=not opts.unicode,
    )


class Renderer(lang.Abstract):
    def __init__(self, opts: RenderingOptions) -> None:
        super().__init__()
        self._opts = opts
        self._kw = make_render_kwargs(opts)


##


def term_color(o: ta.Any, state: JsonRenderer.State) -> tuple[str, str]:
    if state is JsonRenderer.State.KEY:
        return term.SGR(term.SGRs.FG.BRIGHT_BLUE), term.SGR(term.SGRs.RESET)
    elif isinstance(o, str):
        return term.SGR(term.SGRs.FG.GREEN), term.SGR(term.SGRs.RESET)
    else:
        return '', ''


##


class EagerRenderer(Renderer):
    def render(self, v: ta.Any) -> str:
        if self._opts.raw:
            if not isinstance(v, str):
                raise TypeError(f'Raw output must be strings, got {type(v)}', v)

            return v

        elif self._opts.color:
            return JsonRenderer.render_str(
                v,
                **self._kw,
                style=term_color,
            )

        else:
            return json.dumps(
                v,
                **self._kw,
            )


##


class StreamRenderer(Renderer):
    def __init__(self, opts: RenderingOptions) -> None:
        super().__init__(opts)

        self._renderer = StreamJsonRenderer(
            style=term_color if self._opts.color else None,
            delimiter='\n',
            **self._kw,
        )

    def render(self, e: JsonStreamParserEvent) -> ta.Generator[str, None, None]:
        return self._renderer.render((e,))



"""
TODO:
 - point validate / check exceptions to lambdas
"""
import dataclasses as dc
import types
import typing as ta

from ... import check as check_
from ... import lang
from .exceptions import FieldValidationError
from .internals import FIELDS_ATTR
from .internals import FieldType
from .internals import is_classvar
from .internals import is_initvar
from .internals import is_kw_only
from .params import get_field_extras
from .processing import Processor


if ta.TYPE_CHECKING:
    from . import api
else:
    api = lang.proxy_import('.api', __package__)


MISSING = dc.MISSING


##


def raise_field_validation_error(
        obj: ta.Any,
        field: str,
        fn: ta.Callable,
        value: ta.Any,
) -> ta.NoReturn:
    raise FieldValidationError(
        obj,
        field,
        fn,
        value,
    )


##


def field_type(f: dc.Field) -> FieldType:
    if (ft := getattr(f, '_field_type')) is not None:
        return FieldType(ft)
    else:
        return FieldType.INSTANCE


def has_default(f: dc.Field) -> bool:
    return not (f.default is MISSING and f.default_factory is MISSING)


##


class FieldsProcessor(Processor):
    def _process(self) -> None:
        cls = self._info.cls
        fields: dict[str, dc.Field] = {}

        for b in cls.__mro__[-1:0:-1]:
            base_fields = getattr(b, FIELDS_ATTR, None)
            if base_fields is not None:
                for f in base_fields.values():
                    fields[f.name] = f

        cls_fields: list[dc.Field] = []

        kw_only = self._info.params.kw_only
        kw_only_seen = False
        for name, ann in self._info.cls_annotations.items():
            if is_kw_only(cls, ann):
                if kw_only_seen:
                    raise TypeError(f'{name!r} is KW_ONLY, but KW_ONLY has already been specified')
                kw_only_seen = True
                kw_only = True
            else:
                cls_fields.append(preprocess_field(cls, name, ann, kw_only))

        for f in cls_fields:
            fields[f.name] = f
            if isinstance(getattr(cls, f.name, None), dc.Field):
                if f.default is MISSING:
                    delattr(cls, f.name)
                else:
                    setattr(cls, f.name, f.default)

        for name, value in cls.__dict__.items():
            if isinstance(value, dc.Field) and name not in self._info.cls_annotations:
                raise TypeError(f'{name!r} is a field but has no type annotation')

        setattr(cls, FIELDS_ATTR, fields)


##


def preprocess_field(
        cls: type,
        a_name: str,
        a_type: ta.Any,
        default_kw_only: bool,
) -> dc.Field:
    default = getattr(cls, a_name, MISSING)
    if isinstance(default, dc.Field):
        f = default
    else:
        if isinstance(default, types.MemberDescriptorType):
            # This is a field in __slots__, so it has no default value.
            default = MISSING
        f = api.field(default=default)

    f.name = a_name
    f.type = a_type

    ft = FieldType.INSTANCE
    if is_classvar(cls, f.type):
        ft = FieldType.CLASS
    if is_initvar(cls, f.type):
        ft = FieldType.INIT
    if ft in (FieldType.CLASS, FieldType.INIT):
        if f.default_factory is not MISSING:
            raise TypeError(f'field {f.name} cannot have a default factory')
    f._field_type = ft.value  # type: ignore  # noqa

    if ft in (FieldType.INSTANCE, FieldType.INIT):
        if f.kw_only is MISSING:
            f.kw_only = default_kw_only
    else:
        check_.arg(ft is FieldType.CLASS)
        if f.kw_only is not MISSING:
            raise TypeError(f'field {f.name} is a ClassVar but specifies kw_only')

    if ft is FieldType.INSTANCE and f.default is not MISSING and f.default.__class__.__hash__ is None:
        raise ValueError(f'mutable default {type(f.default)} for field {f.name} is not allowed: use default_factory')

    return f


def field_assign(
        frozen: bool,
        name: str,
        value: ta.Any,
        self_name: str,
        override: bool,
) -> str:
    if override:
        return f'{self_name}.__dict__[{name!r}] = {value}'
    if frozen:
        return f'__dataclass_builtins_object__.__setattr__({self_name}, {name!r}, {value})'
    return f'{self_name}.{name} = {value}'


def field_init(
        f: dc.Field,
        frozen: bool,
        locals: dict[str, ta.Any],  # noqa
        self_name: str,
        slots: bool,
) -> ta.Sequence[str]:
    default_name = f'__dataclass_dflt_{f.name}__'
    fx = get_field_extras(f)

    lines = []

    value: str | None = None
    if f.default_factory is not MISSING:
        if f.init:
            locals[default_name] = f.default_factory
            lines.append(f'if {f.name} is __dataclass_HAS_DEFAULT_FACTORY__: {f.name} = {default_name}()')
            value = f.name
        else:
            locals[default_name] = f.default_factory
            lines.append(f'{f.name} = {default_name}()')
            value = f.name

    elif f.init:
        if f.default is MISSING:
            value = f.name
        elif f.default is not MISSING:
            locals[default_name] = f.default  # Not referenced her, just useful / consistent to have in function scope
            value = f.name

    elif slots and f.default is not MISSING:
        locals[default_name] = f.default
        lines.append(f'{f.name} = {default_name}')
        value = default_name

    else:
        pass

    if fx.derive is not None:
        raise NotImplementedError

    if fx.frozen:
        raise NotImplementedError

    if fx.coerce is not None:
        cn = f'__dataclass_coerce__{f.name}__'
        locals[cn] = fx.coerce
        lines.append(f'{value} = {cn}({value})')

    if fx.validate is not None:
        cn = f'__dataclass_validate__{f.name}__'
        locals[cn] = fx.validate
        lines.append(
            f'if not {cn}({value}): '
            f'__dataclass_raise_field_validation_error__({self_name}, {f.name!r}, {cn}, {value})',
        )

    if fx.check_type:
        cn = f'__dataclass_check_type__{f.name}__'
        ct: ta.Any
        if isinstance(fx.check_type, tuple):
            ct = tuple(type(None) if e is None else check_.isinstance(e, type) for e in fx.check_type)
        elif isinstance(fx.check_type, (type, tuple)):
            ct = fx.check_type
        # FIXME:
        # elif info.params_extras.generic_init:
        #     ct = info.generic_replaced_field_annotations[f.name]
        else:
            ct = f.type
        locals[cn] = ct
        lines.append(
            f'if not __dataclass_builtins_isinstance__({value}, {cn}): '
            f'raise __dataclass_builtins_TypeError__({value}, {cn})',
        )

    if value is not None and field_type(f) is not FieldType.INIT:
        lines.append(field_assign(frozen, f.name, value, self_name, fx.override))  # noqa

    return lines



import dataclasses as dc
import typing as ta

from .... import lang


if ta.TYPE_CHECKING:
    from ....specs import jmespath
else:
    jmespath = lang.proxy_import('....specs.jmespath', __package__)


##


@dc.dataclass(frozen=True, kw_only=True)
class ProcessingOptions:
    jmespath_expr: ta.Any | None = None
    flat: bool = False


class Processor:
    def __init__(self, opts: ProcessingOptions) -> None:
        super().__init__()

        self._opts = opts

        jmespath_expr = opts.jmespath_expr
        if isinstance(jmespath_expr, str):
            jmespath_expr = jmespath.compile(jmespath_expr)
        self._jmespath_expr: ta.Any | None = jmespath_expr

    def process(self, v: ta.Any) -> ta.Iterable[ta.Any]:
        if self._jmespath_expr is not None:
            v = self._jmespath_expr.search(v)

        if self._opts.flat:
            if (
                    not isinstance(v, ta.Iterable) or  # noqa
                    isinstance(v, ta.Mapping) or
                    isinstance(v, lang.BUILTIN_SCALAR_ITERABLE_TYPES)
            ):
                raise TypeError(f'Flat output must be flat collections, got {type(v)}', v)

            yield from v

        else:
            yield v



"""
TODO:
 - select best debugger: ipdb, pudb,
"""
import contextlib
import sys
import traceback
import typing as ta


@contextlib.contextmanager
def debugging_on_exception(
        *,
        silent: bool = False,
) -> ta.Iterator[None]:
    import pdb  # noqa

    try:
        yield
    except Exception:
        if not silent:
            traceback.print_exc()

        pdb.post_mortem(sys.exc_info()[2])

        raise



import typing as ta


T = ta.TypeVar('T')


class _Proxy(ta.Generic[T]):
    __obj: T | None = None

    @property
    def _obj(self) -> T:
        if self.__obj is None:
            raise TypeError('recursive proxy not set')
        return self.__obj

    def _set_obj(self, obj: T) -> None:
        if self.__obj is not None:
            raise TypeError('recursive proxy already set')
        self.__obj = obj

    @classmethod
    def _new(cls) -> tuple[ta.Any, ta.Callable[[ta.Any], None]]:
        return (p := cls()), p._set_obj  # noqa



# ruff: noqa: I001
import typing as _ta

from ... import lang as _lang


from .consts import (  # noqa
    COMPACT_KWARGS,
    COMPACT_SEPARATORS,
    PRETTY_INDENT,
    PRETTY_KWARGS,
    PRETTY_SEPARATORS,
    Separators,
)

from .encoding import (  # noqa
    decodes,
    detect_encoding,
)

from .json import (  # noqa
    dump,
    dump_compact,
    dump_pretty,
    dumps,
    dumps_compact,
    dumps_pretty,
    load,
    loads,
)

if _ta.TYPE_CHECKING:
    from .render import (  # noqa
        JsonRenderer,
    )
else:
    _lang.proxy_init(globals(), '.render', [
        'JsonRenderer',
    ])



# ruff: noqa: UP007
import io
import typing as ta

from .check import check_isinstance
from .check import check_not_none
from .strings import attr_repr


class DelimitingBuffer:
    """
    https://github.com/python-trio/trio/issues/796 :|
    """

    #

    class Error(Exception):
        def __init__(self, buffer: 'DelimitingBuffer') -> None:
            super().__init__(buffer)
            self.buffer = buffer

        def __repr__(self) -> str:
            return attr_repr(self, 'buffer')

    class ClosedError(Error):
        pass

    #

    DEFAULT_DELIMITERS: bytes = b'\n'

    def __init__(
            self,
            delimiters: ta.Iterable[int] = DEFAULT_DELIMITERS,
            *,
            keep_ends: bool = False,
            max_size: ta.Optional[int] = None,
    ) -> None:
        super().__init__()

        self._delimiters = frozenset(check_isinstance(d, int) for d in delimiters)
        self._keep_ends = keep_ends
        self._max_size = max_size

        self._buf: ta.Optional[io.BytesIO] = io.BytesIO()

    #

    @property
    def is_closed(self) -> bool:
        return self._buf is None

    def tell(self) -> int:
        if (buf := self._buf) is None:
            raise self.ClosedError(self)
        return buf.tell()

    def peek(self) -> bytes:
        if (buf := self._buf) is None:
            raise self.ClosedError(self)
        return buf.getvalue()

    def _find_delim(self, data: ta.Union[bytes, bytearray], i: int) -> ta.Optional[int]:
        r = None  # type: int | None
        for d in self._delimiters:
            if (p := data.find(d, i)) >= 0:
                if r is None or p < r:
                    r = p
        return r

    def _append_and_reset(self, chunk: bytes) -> bytes:
        buf = check_not_none(self._buf)
        if not buf.tell():
            return chunk

        buf.write(chunk)
        ret = buf.getvalue()
        buf.seek(0)
        buf.truncate()
        return ret

    class Incomplete(ta.NamedTuple):
        b: bytes

    def feed(self, data: ta.Union[bytes, bytearray]) -> ta.Generator[ta.Union[bytes, Incomplete], None, None]:
        if (buf := self._buf) is None:
            raise self.ClosedError(self)

        if not data:
            self._buf = None

            if buf.tell():
                yield self.Incomplete(buf.getvalue())

            return

        l = len(data)
        i = 0
        while i < l:
            if (p := self._find_delim(data, i)) is None:
                break

            n = p + 1
            if self._keep_ends:
                p = n

            yield self._append_and_reset(data[i:p])

            i = n

        if i >= l:
            return

        if self._max_size is None:
            buf.write(data[i:])
            return

        while i < l:
            remaining_data_len = l - i
            remaining_buf_capacity = self._max_size - buf.tell()

            if remaining_data_len < remaining_buf_capacity:
                buf.write(data[i:])
                return

            p = i + remaining_buf_capacity
            yield self.Incomplete(self._append_and_reset(data[i:p]))
            i = p



import inspect
import os.path
import subprocess
import tempfile

from ... import argparse as ap
from .pycharm import get_pycharm_version


class Cli(ap.Cli):
    @ap.command()
    def version(self) -> None:
        print(get_pycharm_version())

    @ap.command(
        ap.arg('python-exe'),
        ap.arg('args', nargs=ap.REMAINDER),
    )
    def runhack(self) -> int:
        if not os.path.isfile(exe := self.args.python_exe):
            raise FileNotFoundError(exe)

        from .._pycharm import runhack
        src = inspect.getsource(runhack)

        src_file = tempfile.mktemp(__package__ + '-runhack')  # noqa
        with open(src_file, 'w') as f:
            f.write(src)

        proc = subprocess.run([exe, src_file, *self.args.args], check=False)
        return proc.returncode


def _main() -> None:
    Cli().call_and_exit()


# @omlish-manifest
_CLI_MODULE = {'$omdev.cli.types.CliModule': {
    'cmd_name': 'pycharm',
    'mod_name': __name__,
}}


if __name__ == '__main__':
    _main()



"""
A small but likely growing collection of (completely optional) tools to make pydevd (PyCharm's, among other python
IDE's, debugger) do hard things. Originally explored and added to get spark jvm python subprocesses to connect back to
an already-debugging PyCharm instance to debug PySpark jobs.

TODO:
 - https://www.jetbrains.com/help/pycharm/remote-debugging-with-product.html#
 - PyCharm.app/Contents/plugins/python-ce/helpers/pydev/_pydevd_bundle/pydevd_constants.py -> USE_LOW_IMPACT_MONITORING

==

https://www.jetbrains.com/help/pycharm/remote-debugging-with-product.html#remote-debug-config ->

pycharm_port = 43251
pycharm_version = '241.18034.82'
buf = textwrap.dedent(f'''
    import subprocess
    import sys
    subprocess.check_call([sys.executable, '-mpip', 'install', f'pydevd-pycharm~={pycharm_version}'])

    import pydevd_pycharm  # noqa
    pydevd_pycharm.settrace(
        'docker.for.mac.localhost',
         port={pycharm_port},
          stdoutToServer=True,
           stderrToServer=True,
       )
''') + '\n' * 2 + buf
"""
import json
import os
import sys
import tempfile
import textwrap
import threading
import types
import typing as ta

from .. import check
from .. import lang


ALLOW_DEBUGGER_CALLS = False


DEBUGGER_CALL_PACKAGES = {
    '_pydevd_bundle',
}


def is_debugger_call(hoist: int = 0, walk: int = 2) -> bool:
    frame: ta.Optional[types.FrameType] = sys._getframe(2 + hoist)  # noqa
    for _ in range(walk):
        if frame is None:
            break
        package = frame.f_globals.get('__package__')
        if package in DEBUGGER_CALL_PACKAGES:
            return True
        frame = frame.f_back
    return False


class DebuggerCallForbiddenError(Exception):
    pass


def forbid_debugger_call(hoist: int = 0) -> None:
    # FIXME: only reentrant?
    if not ALLOW_DEBUGGER_CALLS and is_debugger_call(hoist + 1):
        raise DebuggerCallForbiddenError


##


@lang.cached_function
def silence_subprocess_check() -> None:
    try:
        # /Applications/PyCharm.app/Contents/plugins/python/helpers/pydev/_pydev_bundle/pydev_monkey.py
        from _pydev_bundle import pydev_monkey  # noqa
    except ImportError:
        return

    new_tb = lang.proxy_import('traceback')
    new_tb.print_exc = lambda *a, **k: None  # type: ignore  # noqa
    pydev_monkey.traceback = new_tb


@lang.cached_function
def patch_for_trio_asyncio() -> None:
    """Fix for `trio a callable object was expected by call_soon(), got Task`"""

    try:
        import pydevd_nest_asyncio  # noqa
    except ImportError:
        return

    import trio_asyncio._base  # noqa

    def new_call_soon(self, callback, *args, **context):
        _, callback = pydevd_nest_asyncio._PydevdAsyncioUtils.try_to_get_internal_callback(callback)  # noqa
        return orig_call_soon(self, callback, *args, **context)

    orig_call_soon = trio_asyncio._base.BaseTrioEventLoop.call_soon  # noqa
    trio_asyncio._base.BaseTrioEventLoop.call_soon = new_call_soon  # noqa


##


@lang.cached_function
def _pydevd() -> types.ModuleType | None:
    try:
        return __import__('pydevd')
    except ImportError:
        return None


def is_present() -> bool:
    return _pydevd() is not None


def get_setup() -> dict | None:
    if is_present():
        return _pydevd().SetupHolder.setup
    else:
        return None


def is_running() -> bool:
    return get_setup() is not None


##


ARGS_ENV_VAR = 'PYDEVD_ARGS'


def get_args() -> list[str]:
    check.state(is_present())
    setup: ta.Mapping[ta.Any, ta.Any] = check.isinstance(get_setup(), dict)
    args = [_pydevd().__file__]

    for k in [
        'port',
        'vm_type',
        'client',
    ]:
        if v := setup[k]:
            args.extend(['--' + k, str(v)])

    for k in [
        'server',
        'multiproc',
        'multiprocess',
        'save-signatures',
        'save-threading',
        'save-asyncio',
        'print-in-debugger-startup',
        'cmd-line',
    ]:
        if setup[k]:
            args.append('--' + k)

    if setup['qt-support']:
        args.append('--qt-support=' + setup['qt-support'])

    return args


def save_args() -> None:
    if is_present():
        os.environ[ARGS_ENV_VAR] = json.dumps(get_args())


def maybe_reexec(
        *,
        file: str | None = None,
        module: str | None = None,
        silence: bool = False,
) -> None:
    if ARGS_ENV_VAR not in os.environ:
        return

    try:
        import pydevd  # noqa
    except ImportError:
        return

    if pydevd.SetupHolder.setup is not None:  # noqa
        return

    if module is not None:
        if file is not None:
            raise ValueError

        tmpdir = tempfile.mkdtemp()
        bootstrap_path = os.path.join(tmpdir, 'bootstrap.py')
        with open(bootstrap_path, 'w') as f:
            f.write(textwrap.dedent(f"""
            import sys
            old_paths = set(sys.path)
            for new_path in {sys.path!r}:
                if new_path not in old_paths:
                    sys.path.insert(0, new_path)

            import runpy
            runpy.run_module({module!r}, run_name='__main__')
            """))
        file = bootstrap_path

    elif file is None:
        raise ValueError

    args = [sys.executable]
    args.extend(json.loads(os.environ[ARGS_ENV_VAR]))
    args.extend(['--file', file])
    args.extend(sys.argv[1:])

    if silence:
        tmpdir = tempfile.mkdtemp()
        bootstrap_path = os.path.join(tmpdir, 'bootstrap.py')
        with open(bootstrap_path, 'w') as f:
            f.write(textwrap.dedent(f"""
            import sys
            old_paths = set(sys.path)
            for new_path in {sys.path!r}:
                if new_path not in old_paths:
                    sys.path.insert(0, new_path)

            _stderr_write = sys.stderr.write
            def stderr_write(*args, **kwargs):
                code = sys._getframe(1).f_code
                if code is not None and code.co_filename and code.co_filename.endswith('/pydev_log.py'):
                    return
                _stderr_write(*args, **kwargs)
            sys.stderr.write = stderr_write

            sys.argv = {args[1:]!r}
            import runpy
            runpy.run_path({args[1]!r}, run_name='__main__')
            """))
        args = [args[0], bootstrap_path]

    os.execvp(sys.executable, args)


def debug_unhandled_exception(exc_info: ta.Any = None) -> None:
    if exc_info is None:
        exc_info = sys.exc_info()

    try:
        import pydevd
        from pydevd import pydevd_tracing

    except ImportError:
        return

    exctype, value, traceback = exc_info
    frames = []
    while traceback:
        frames.append(traceback.tb_frame)
        traceback = traceback.tb_next

    thread = threading.current_thread()
    frames_by_id = {id(frame): frame for frame in frames}
    frame = frames[-1]
    exception = (exctype, value, traceback)

    if hasattr(thread, 'additional_info'):
        thread.additional_info.pydev_message = 'server exception'
    try:
        debugger = pydevd.debugger  # noqa
    except AttributeError:
        debugger = pydevd.get_global_debugger()  # noqa

    pydevd_tracing.SetTrace(None)

    debugger.stop_on_unhandled_exception(thread, frame, frames_by_id, exception)



import abc
import functools
import threading
import typing as ta

from .restrict import Final
from .restrict import NotInstantiable


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class SimpleMetaDict(dict):
    def update(self, m: ta.Mapping[K, V], **kwargs: V) -> None:  # type: ignore
        for k, v in m.items():
            self[k] = v
        for k, v in kwargs.items():  # type: ignore
            self[k] = v


##


class _NamespaceMeta(abc.ABCMeta):
    def __new__(mcls, name, bases, namespace):
        if bases:
            for nc in (NotInstantiable,):
                if nc not in bases:
                    bases += (nc,)
        return super().__new__(mcls, name, bases, namespace)

    def __iter__(cls) -> ta.Iterator[tuple[str, ta.Any]]:
        seen: set[str] = set()
        for bcls in reversed(cls.__mro__):
            for a in bcls.__dict__:
                if a in seen:
                    continue
                seen.add(a)
                if not a.startswith('_'):
                    yield (a, getattr(cls, a))

    def __getitem__(cls, n: str) -> ta.Any:
        return getattr(cls, n)


class Namespace(metaclass=_NamespaceMeta):
    pass


##


_MARKER_NAMESPACE_KEYS: set[str] | None = None


class _MarkerMeta(abc.ABCMeta):
    def __new__(mcls, name, bases, namespace):
        global _MARKER_NAMESPACE_KEYS

        if _MARKER_NAMESPACE_KEYS is None:
            if not (namespace.get('__module__') == __name__ and name == 'Marker'):
                raise RuntimeError
            _MARKER_NAMESPACE_KEYS = set(namespace)

        else:
            if set(namespace) - _MARKER_NAMESPACE_KEYS:
                raise TypeError('Markers must not include contents. Did you mean to use Namespace?')
            if Final not in bases:
                bases += (Final,)

        return super().__new__(mcls, name, bases, namespace)

    def __instancecheck__(self, instance):
        return instance is self

    def __repr__(cls) -> str:
        return f'<{cls.__name__}>'


class Marker(NotInstantiable, metaclass=_MarkerMeta):
    """A marker."""

    def __bool__(self):
        raise TypeError


##


_SINGLETON_INSTANCE_ATTR = '__singleton_instance__'
_SINGLETON_LOCK = threading.RLock()


def _set_singleton_instance(inst):
    cls = inst.__class__
    if _SINGLETON_INSTANCE_ATTR in cls.__dict__:
        raise TypeError(cls)

    inst.__init__()
    old_init = cls.__init__

    @functools.wraps(old_init)
    def new_init(self):
        if self.__class__ is not cls:
            old_init(self)  # noqa

    setattr(cls, '__init__', new_init)
    setattr(cls, _SINGLETON_INSTANCE_ATTR, inst)

    return inst


class Singleton:
    def __new__(cls):
        return cls.__dict__[_SINGLETON_INSTANCE_ATTR]

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)
        _set_singleton_instance(super().__new__(cls))  # noqa


class LazySingleton:
    def __new__(cls):
        try:
            return cls.__dict__[_SINGLETON_INSTANCE_ATTR]
        except KeyError:
            pass
        with _SINGLETON_LOCK:
            try:
                return cls.__dict__[_SINGLETON_INSTANCE_ATTR]
            except KeyError:
                pass
            return _set_singleton_instance(super().__new__(cls))



# ruff: noqa: UP006
import functools
import typing as ta


T = ta.TypeVar('T')


_GENERIC_ALIAS_TYPES = (
    ta._GenericAlias,  # type: ignore  # noqa
    *([ta._SpecialGenericAlias] if hasattr(ta, '_SpecialGenericAlias') else []),  # noqa
)


def is_generic_alias(obj, *, origin: ta.Any = None) -> bool:
    return (
        isinstance(obj, _GENERIC_ALIAS_TYPES) and
        (origin is None or ta.get_origin(obj) is origin)
    )


is_union_alias = functools.partial(is_generic_alias, origin=ta.Union)
is_callable_alias = functools.partial(is_generic_alias, origin=ta.Callable)


def is_optional_alias(spec: ta.Any) -> bool:
    return (
        isinstance(spec, _GENERIC_ALIAS_TYPES) and  # noqa
        ta.get_origin(spec) is ta.Union and
        len(ta.get_args(spec)) == 2 and
        any(a in (None, type(None)) for a in ta.get_args(spec))
    )


def get_optional_alias_arg(spec: ta.Any) -> ta.Any:
    [it] = [it for it in ta.get_args(spec) if it not in (None, type(None))]
    return it


def deep_subclasses(cls: ta.Type[T]) -> ta.Iterator[ta.Type[T]]:
    seen = set()
    todo = list(reversed(cls.__subclasses__()))
    while todo:
        cur = todo.pop()
        if cur in seen:
            continue
        seen.add(cur)
        yield cur
        todo.extend(reversed(cur.__subclasses__()))



import abc
import dataclasses as dc
import typing as ta

from ... import check
from ... import lang
from .copy import CopyProcessor
from .fields import FieldsProcessor
from .frozen import FrozenProcessor
from .hashing import HashProcessor
from .init import InitProcessor
from .internals import FIELDS_ATTR
from .internals import PARAMS_ATTR
from .internals import Params
from .order import OrderProcessor
from .params import ParamsExtras
from .processing import Processor
from .reflect import ClassInfo
from .replace import ReplaceProcessor
from .repr import ReprProcessor
from .simple import DocProcessor
from .simple import EqProcessor
from .simple import MatchArgsProcessor
from .simple import OverridesProcessor
from .slots import add_slots


MISSING = dc.MISSING


class MainProcessor:
    def __init__(self, cls: type) -> None:
        super().__init__()

        self._cls = check.isinstance(cls, type)
        self._info = info = ClassInfo(cls, _constructing=True)

        check.not_in(FIELDS_ATTR, cls.__dict__)
        check.is_(check.isinstance(cls.__dict__[PARAMS_ATTR], Params), info.params)
        check.is_(check.isinstance(check.not_none(info.cls_metadata)[ParamsExtras], ParamsExtras), info.params_extras)  # noqa

    def _check_params(self) -> None:
        if self._info.params.order and not self._info.params.eq:
            raise ValueError('eq must be true if order is true')

    @lang.cached_function
    def _transform_slots(self) -> None:
        if self._info.params.weakref_slot and not self._info.params.slots:
            raise TypeError('weakref_slot is True but slots is False')
        if not self._info.params.slots:
            return
        self._cls = add_slots(self._cls, self._info.params.frozen, self._info.params.weakref_slot)

    PROCESSOR_TYPES: ta.ClassVar[ta.Sequence[type[Processor]]] = [
        FieldsProcessor,
        InitProcessor,
        OverridesProcessor,
        ReprProcessor,
        EqProcessor,
        OrderProcessor,
        FrozenProcessor,
        HashProcessor,
        DocProcessor,
        MatchArgsProcessor,
        ReplaceProcessor,
        CopyProcessor,
    ]

    @lang.cached_function
    def process(self) -> type:
        self._check_params()

        ps = [pcls(self._info) for pcls in self.PROCESSOR_TYPES]

        for p in ps:
            p.check()

        for p in ps:
            p.process()

        self._transform_slots()

        abc.update_abstractmethods(self._cls)  # noqa

        return self._cls


def process_class(cls: type) -> type:
    return MainProcessor(cls).process()



from .abstract import (  # noqa
    Abstract,
    is_abstract,
    is_abstract_class,
    is_abstract_method,
    make_abstract,
    unabstract_class,
)

from .restrict import (  # noqa
    AnySensitive,
    Final,
    FinalError,
    NoBool,
    NotInstantiable,
    NotPicklable,
    PackageSealed,
    SENSITIVE_ATTR,
    Sealed,
    SealedError,
    Sensitive,
    no_bool,
)

from .simple import (  # noqa
    LazySingleton,
    Marker,
    Namespace,
    SimpleMetaDict,
    Singleton,
)

from .virtual import (  # noqa
    Callable,
    Descriptor,
    Picklable,
    Virtual,
    virtual_check,
)



import abc
import contextlib
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .. import reflect as rfl
from .bindings import Binding
from .elements import Element
from .keys import Key
from .keys import as_key
from .providers import Provider
from .types import Scope


if ta.TYPE_CHECKING:
    from . import injector as injector_
else:
    injector_ = lang.proxy_import('.injector', __package__)


##


SCOPE_ALIASES: dict[str, Scope] = {}


@dc.dataclass(frozen=True)
@dc.extra_params(cache_hash=True)
class ScopeBinding(Element, lang.Final):
    scope: Scope = dc.xfield(coerce=check.of_isinstance(Scope))


def bind_scope(sc: Scope) -> Element:
    return ScopeBinding(sc)


##


class Singleton(Scope, lang.Singleton, lang.Final):
    pass


SCOPE_ALIASES['singleton'] = Singleton()


##


class Thread(Scope, lang.Singleton, lang.Final):
    pass


SCOPE_ALIASES['thread'] = Thread()


##


@dc.dataclass(frozen=True)
@dc.extra_params(cache_hash=True)
class SeededScope(Scope, lang.Final):
    tag: ta.Any = dc.xfield(coerce=check.not_none)

    class Manager(lang.Abstract):
        @abc.abstractmethod
        def __call__(self, seeds: ta.Mapping[Key, ta.Any]) -> ta.ContextManager[None]:
            raise NotImplementedError


@dc.dataclass(frozen=True)
@dc.extra_params(cache_hash=True)
class ScopeSeededProvider(Provider):
    ss: SeededScope = dc.xfield(coerce=check.of_isinstance(SeededScope))
    key: Key = dc.xfield(coerce=check.of_isinstance(Key))

    def provided_ty(self) -> rfl.Type | None:
        return self.key.ty


def bind_scope_seed(ss: SeededScope, k: ta.Any) -> Element:
    k = as_key(k)
    return Binding(k, ScopeSeededProvider(ss, k))


@contextlib.contextmanager
def enter_seeded_scope(
        i: injector_.Injector,
        ss: SeededScope,
        keys: ta.Mapping[Key, ta.Any],
) -> ta.Generator[None, None, None]:
    with i.provide(Key(SeededScope.Manager, tag=ss))(keys):
        yield



import abc
import dataclasses as dc
import typing as ta

from ... import defs


class AbstractFieldDescriptor(abc.ABC):

    def __init__(
            self,
            *,
            default: ta.Any = dc.MISSING,
            frozen: bool = False,
            name: str | None = None,
            pre_set: ta.Callable[[ta.Any, ta.Any], ta.Any] | None = None,
            post_set: ta.Callable[[ta.Any, ta.Any], None] | None = None,
    ) -> None:
        super().__init__()

        self._default = default
        self._frozen = frozen
        self._name = name
        self._pre_set = pre_set
        self._post_set = post_set

    defs.repr('name')
    defs.getter('default', 'frozen', 'name', 'pre_set', 'post_set')

    def __set_name__(self, owner, name):
        if self._name is None:
            self._name = name

    def __get__(self, instance, owner=None):
        if instance is not None:
            try:
                return self._get(instance)
            except AttributeError:
                pass
        if self._default is not dc.MISSING:
            return self._default
        raise AttributeError(self._name)

    @abc.abstractmethod
    def _get(self, instance):
        raise NotImplementedError

    def __set__(self, instance, value):
        if self._frozen:
            raise dc.FrozenInstanceError(f'cannot assign to field {self._name!r}')
        if self._pre_set is not None:
            value = self._pre_set(instance, value)
        self._set(instance, value)
        if self._post_set is not None:
            self._post_set(instance, value)

    @abc.abstractmethod
    def _set(self, instance, value):
        raise NotImplementedError

    def __delete__(self, instance):
        if self._frozen:
            raise dc.FrozenInstanceError(f'cannot delete field {self._name!r}')
        self._del(instance)

    @abc.abstractmethod
    def _del(self, instance):
        raise NotImplementedError


class PyFieldDescriptor(AbstractFieldDescriptor):

    def __init__(
            self,
            attr: str,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__(**kwargs)

        self._attr = attr

    defs.repr('attr', 'name')
    defs.getter('attr')

    def _get(self, instance):
        return getattr(instance, self._attr)

    def _set(self, instance, value):
        setattr(instance, self._attr, value)

    def _del(self, instance):
        delattr(instance, self._attr)


FieldDescriptor = PyFieldDescriptor



"""
TODO:
 - read from http
 - jmespath output flat, unquoted strs like jq '.[]'

==

jq Command options:
  -n, --null-input          use `null` as the single input value;
  -R, --raw-input           read each line as string instead of JSON;
  -s, --slurp               read all inputs into an array and use it as the single input value;
  -c, --compact-output      compact instead of pretty-printed output;
  -r, --raw-output          output strings without escapes and quotes;
      --raw-output0         implies -r and output NUL after each output;
  -j, --join-output         implies -r and output without newline after each output;
  -a, --ascii-output        output strings by only ASCII characters using escape sequences;
  -S, --sort-keys           sort keys of each object on output;
  -C, --color-output        colorize JSON output;
  -M, --monochrome-output   disable colored output;
      --tab                 use tabs for indentation;
      --indent n            use n spaces for indentation (max 7 spaces);
      --unbuffered          flush output stream after each output;
      --stream              parse the input value in streaming fashion;
      --stream-errors       implies --stream and report parse error as an array;
      --seq                 parse input/output as application/json-seq;
  -f, --from-file file      load filter from the file;
  -L directory              search modules from the directory;
      --arg name value      set $name to the string value;
      --argjson name value  set $name to the JSON value;
      --slurpfile name file set $name to an array of JSON values read from the file;
      --rawfile name file   set $name to string contents of file;
      --args                consume remaining arguments as positional string values;
      --jsonargs            consume remaining arguments as positional JSON values;
  -e, --exit-status         set exit status code based on the output;
  -V, --version             show the version;
  --build-configuration     show jq's build configuration;
  -h, --help                show the help;
  --                        terminates argument processing;
"""
import argparse
import contextlib
import dataclasses as dc
import io
import os
import subprocess
import sys
import typing as ta

from .... import check
from .... import fnpipes as fp
from .... import lang
from .formats import FORMATS_BY_NAME
from .formats import Format
from .formats import Formats
from .parsing import DelimitingParser
from .parsing import EagerParser
from .parsing import StreamBuilder
from .parsing import StreamParser
from .processing import ProcessingOptions
from .processing import Processor
from .rendering import EagerRenderer
from .rendering import RenderingOptions
from .rendering import StreamRenderer


T = ta.TypeVar('T')
U = ta.TypeVar('U')


def _build_args_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()

    parser.add_argument('file', nargs='?')

    parser.add_argument('--stream', action='store_true')
    parser.add_argument('--stream-build', action='store_true')

    parser.add_argument('-l', '--lines', action='store_true')

    parser.add_argument('--read-buffer-size', type=int, default=0x4000)

    parser.add_argument('-f', '--format')

    parser.add_argument('-x', '--jmespath-expr')
    parser.add_argument('-F', '--flat', action='store_true')

    parser.add_argument('-z', '--compact', action='store_true')
    parser.add_argument('-p', '--pretty', action='store_true')
    parser.add_argument('-i', '--indent')
    parser.add_argument('-s', '--sort-keys', action='store_true')
    parser.add_argument('-R', '--raw', action='store_true')
    parser.add_argument('-U', '--unicode', action='store_true')
    parser.add_argument('-c', '--color', action='store_true')

    parser.add_argument('-L', '--less', action='store_true')

    return parser


def _parse_args(args: ta.Any = None) -> ta.Any:
    return _build_args_parser().parse_args(args)


@dc.dataclass(frozen=True, kw_only=True)
class RunConfiguration:
    format: Format
    processing: ProcessingOptions
    rendering: RenderingOptions


def _process_args(args: ta.Any) -> RunConfiguration:
    fmt_name = args.format
    if fmt_name is None:
        if args.file is not None:
            ext = args.file.rpartition('.')[2]
            if ext in FORMATS_BY_NAME:
                fmt_name = ext
    if fmt_name is None:
        fmt_name = 'json'
    format = FORMATS_BY_NAME[fmt_name]  # noqa

    if args.stream:
        check.arg(format is Formats.JSON.value)

    #

    processing = ProcessingOptions(
        jmespath_expr=args.jmespath_expr,
        flat=args.flat,
    )

    #

    separators = None
    if args.compact:
        separators = (',', ':')

    indent = None
    if args.pretty:
        indent = 2
    if args.indent:
        try:
            indent = int(args.indent)
        except ValueError:
            indent = args.indent

    rendering = RenderingOptions(
        indent=indent,
        separators=separators,
        sort_keys=args.sort_keys,
        raw=args.raw,
        unicode=args.unicode,
        color=args.color,
    )

    #

    return RunConfiguration(
        format=format,
        processing=processing,
        rendering=rendering,
    )


def _main() -> None:
    args = _parse_args()

    #

    cfg = _process_args(args)

    #

    with contextlib.ExitStack() as es:
        if args.file is None:
            in_file = sys.stdin.buffer

        else:
            in_file = es.enter_context(open(args.file, 'rb'))

        def yield_input() -> ta.Generator[bytes, None, None]:
            fd = check.isinstance(in_file.fileno(), int)

            while True:
                buf = os.read(fd, args.read_buffer_size)

                yield buf

                if not buf:
                    break

        #

        if args.less:
            less = subprocess.Popen(
                [
                    'less',
                    *(['-R'] if cfg.rendering.color else []),
                ],
                stdin=subprocess.PIPE,
                encoding='utf-8',
            )
            out = check.not_none(less.stdin)

            def close_less() -> None:
                out.close()
                less.wait()

            es.enter_context(lang.defer(close_less))  # noqa

        else:
            out = sys.stdout

        #

        parser: ta.Any
        renderer: ta.Any

        if args.stream:
            with contextlib.ExitStack() as es2:
                parser = es2.enter_context(StreamParser())

                def flush_output(
                        fn: ta.Callable[[T], ta.Iterable[U]],
                        i: T,
                ) -> ta.Generator[U, None, None]:
                    n = 0
                    for o in fn(i):
                        yield o
                        n += 1
                    if n:
                        out.flush()

                pipeline: ta.Any

                if args.stream_build:
                    builder: StreamBuilder = es2.enter_context(StreamBuilder())
                    processor = Processor(cfg.processing)
                    renderer = EagerRenderer(cfg.rendering)
                    trailing_newline = False

                    def append_newlines(
                            fn: ta.Callable[[T], ta.Iterable[str]],
                            i: T,
                    ) -> ta.Generator[str, None, None]:
                        yield from fn(i)
                        yield '\n'

                    pipeline = lambda v: (renderer.render(v),)  # Any -> [str]  # noqa
                    pipeline = fp.bind(append_newlines, pipeline)  # Any -> [str]
                    pipeline = fp.bind(lang.flatmap, pipeline)  # [Any] -> [str]
                    pipeline = fp.pipe(fp.bind(lang.flatmap, processor.process), pipeline)  # [Any] -> [str]
                    pipeline = fp.pipe(fp.bind(lang.flatmap, builder.build), pipeline)  # [JsonStreamParserEvent] -> [str]  # noqa
                    pipeline = fp.pipe(parser.parse, pipeline)  # bytes -> [str]

                else:
                    renderer = StreamRenderer(cfg.rendering)
                    trailing_newline = True

                    pipeline = renderer.render  # JsonStreamParserEvent -> [str]
                    pipeline = fp.bind(lang.flatmap, pipeline)  # [JsonStreamParserEvent] -> [str]
                    pipeline = fp.pipe(parser.parse, pipeline)  # bytes -> [str]

                pipeline = fp.bind(flush_output, pipeline)  # bytes -> [str]

                for buf in yield_input():
                    for s in pipeline(buf):
                        print(s, file=out, end='')

                if trailing_newline:
                    print(file=out)

        elif args.lines:
            parser = DelimitingParser(cfg.format)
            processor = Processor(cfg.processing)
            renderer = EagerRenderer(cfg.rendering)

            for buf in yield_input():
                for v in parser.parse(buf):
                    for e in processor.process(v):
                        s = renderer.render(e)
                        print(s, file=out)

        else:
            parser = EagerParser(cfg.format)
            processor = Processor(cfg.processing)
            renderer = EagerRenderer(cfg.rendering)

            with io.TextIOWrapper(in_file) as tf:
                v = parser.parse(tf)

            for e in processor.process(v):
                s = renderer.render(e)
                print(s, file=out)


if __name__ == '__main__':
    _main()



"""
TODO:
 - feed_iter helper
 - accept yielding outputs on transitions, *except* on initial state - add test

See:
 - https://github.com/pytransitions/transitions
"""
import typing as ta


I = ta.TypeVar('I')
O = ta.TypeVar('O')

# MachineGen: ta.TypeAlias = ta.Generator[ta.Iterable[O] | None, I, ta.Optional[MachineGen[I, O]]]
MachineGen: ta.TypeAlias = ta.Generator[ta.Any, ta.Any, ta.Any]


##


class GenMachine(ta.Generic[I, O]):
    """
    Generator-powered state machine. Generators are sent an `I` object and yield any number of `O` objects in response,
    until they yield a `None` by accepting new input. Generators may return a new generator to switch states, or return
    `None` to terminate.
    """

    def __init__(self, initial: MachineGen | None = None) -> None:
        super().__init__()

        if initial is None:
            initial = self._initial_state()
            if initial is None:
                raise TypeError('No initial state')

        self._advance(initial)

    def _initial_state(self) -> MachineGen | None:
        return None

    _gen: MachineGen | None

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}@{hex(id(self))[2:]}<{self.state}>'

    #

    @property
    def state(self) -> str | None:
        if self._gen is not None:
            return self._gen.gi_code.co_qualname
        return None

    #

    @property
    def closed(self) -> bool:
        return self._gen is None

    def close(self) -> None:
        if self._gen is not None:
            self._gen.close()
            self._gen = None

    def __enter__(self) -> ta.Self:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.close()

    #

    class Error(Exception):
        pass

    class ClosedError(Exception):
        pass

    class StateError(Exception):
        pass

    #

    def _advance(self, gen: MachineGen) -> None:
        self._gen = gen

        if (n := next(self._gen)) is not None:  # noqa
            raise GenMachine.ClosedError

    def __call__(self, i: I) -> ta.Iterable[O]:
        if self._gen is None:
            raise GenMachine.ClosedError

        try:
            while (o := self._gen.send(i)) is not None:
                yield from o

        except StopIteration as s:
            if s.value is None:
                self._gen = None
                return None

            self._advance(s.value)



"""
TODO:
 - order preserving / linked list?
 - ItemsView
"""
import abc
import dataclasses as dc
import typing as ta

from .. import lang
from .mappings import yield_dict_init


K = ta.TypeVar('K')
V = ta.TypeVar('V')
K2 = ta.TypeVar('K2')
V2 = ta.TypeVar('V2')


class HashEq(lang.Abstract, ta.Generic[K]):
    @abc.abstractmethod
    def hash(self, k: K) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def eq(self, l: K, r: K) -> bool:
        raise NotImplementedError


@lang.unabstract_class([
    ('hash', '_hash'),
    ('eq', '_eq'),
])
@dc.dataclass(frozen=True)
class HashEq_(ta.Generic[K]):  # noqa
    hash: ta.Callable[[K], int]
    eq: ta.Callable[[K, K], bool]

    def _hash(self, k: K) -> int:
        return self.hash(k)

    def _eq(self, l: K, r: K) -> bool:
        return self.eq(l, r)


class hash_eq(HashEq[K], lang.NotInstantiable, lang.Final):  # noqa
    """Workaround for PEP 695 support."""

    def __new__(  # type: ignore[misc]
            cls,
            hash: ta.Callable[[K], int],  # noqa
            eq: ta.Callable[[K, K], bool],
    ) -> HashEq[K]:
        return HashEq_(hash, eq)  # type: ignore

    def hash(self, k: K) -> int:
        raise TypeError

    def eq(self, l: K, r: K) -> bool:
        raise TypeError


class HashEqMap(ta.MutableMapping[K, V]):
    class _Node(ta.NamedTuple, ta.Generic[K2, V2]):
        k: K2
        v: V2
        h: int

    def __init__(self, hash_eq: HashEq[K], *args: ta.Any, **kwargs: ta.Any) -> None:
        super().__init__()

        self._hash_eq = hash_eq

        self._dct: dict[int, list[HashEqMap._Node[K, V]]] = {}
        self._len = 0

        for k, v in yield_dict_init(*args, **kwargs):
            self[k] = v

    def __len__(self) -> int:
        return self._len

    def clear(self) -> None:
        self._dct.clear()
        self._len = 0

    def __contains__(self, k: K) -> bool:  # type: ignore[override]
        h = self._hash_eq.hash(k)
        try:
            l = self._dct[h]
        except KeyError:
            return False
        return any(self._hash_eq.eq(k, e.k) for e in l)

    def __getitem__(self, k: K) -> V:
        h = self._hash_eq.hash(k)
        l = self._dct[h]
        for e in l:
            if self._hash_eq.eq(k, e.k):
                return e.v
        raise KeyError(k)

    def __setitem__(self, k: K, v: V) -> None:
        h = self._hash_eq.hash(k)
        n = HashEqMap._Node(k, v, h)

        try:
            l = self._dct[h]
        except KeyError:
            l = [n]
            self._dct[h] = l
            self._len += 1
            return

        for i, e in enumerate(l):
            if self._hash_eq.eq(k, e.k):
                l[i] = n
                return

        l.append(n)
        self._len += 1

    def __delitem__(self, k: K) -> None:
        h = self._hash_eq.hash(k)
        l = self._dct[h]
        for i, e in enumerate(l):
            if self._hash_eq.eq(k, e.k):
                del l[i]
                self._len -= 1
                return
        raise KeyError(k)

    def __iter__(self) -> ta.Iterator[K]:
        return self.iterkeys()

    def iterkeys(self) -> ta.Iterator[K]:
        for l in self._dct.values():
            for e in l:
                yield e.k

    def itervalues(self) -> ta.Iterator[V]:
        for l in self._dct.values():
            for e in l:
                yield e.v

    def iteritems(self) -> ta.Iterator[tuple[K, V]]:
        for l in self._dct.values():
            for e in l:
                yield (e.k, e.v)

    # FIXME:

    def keys(self) -> ta.Iterable[K]:  # type: ignore[override]
        return lang.itergen(self.iterkeys)

    def values(self) -> ta.Iterable[V]:  # type: ignore[override]
        return lang.itergen(self.itervalues)

    def items(self) -> ta.Iterable[tuple[K, V]]:  # type: ignore[override]
        return lang.itergen(self.iteritems)



import concurrent.futures as cf
import time
import typing as ta


T = ta.TypeVar('T')


##


class FutureError(Exception, ta.Generic[T]):

    def __init__(self, future: cf.Future, target: T | None = None) -> None:
        super().__init__()

        self._future = future
        self._target = target

    @property
    def future(self) -> cf.Future:
        return self._future

    @property
    def target(self) -> T | None:
        return self._target

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__qualname__}('
            f'exception={self._future.exception()!r}, '
            f'future={self._future!r}, '
            f'target={self._target})'
        )

    __str__ = __repr__


class FutureTimeoutError(Exception):
    pass


def wait_futures(
        futures: ta.Sequence[cf.Future],
        *,
        timeout_s: float = 60,
        tick_interval_s: float = .5,
        tick_fn: ta.Callable[..., bool] = lambda: True,
        raise_exceptions: bool = False,
        cancel_on_exception: bool = False,
) -> bool:
    start = time.time()

    not_done = set(futures)
    while tick_fn():
        done = {f for f in not_done if f.done()}
        if raise_exceptions:
            for fut in done:
                if fut.exception():
                    if cancel_on_exception:
                        for cancel_fut in not_done:
                            cancel_fut.cancel()
                    raise FutureError(fut) from fut.exception()

        not_done -= done
        if not not_done:
            return True

        if time.time() >= (start + timeout_s):
            raise FutureTimeoutError
        time.sleep(tick_interval_s)

    return False


def wait_dependent_futures(
        executor: cf.Executor,
        dependency_sets_by_fn: ta.Mapping[ta.Callable, ta.AbstractSet[ta.Callable]],
        *,
        timeout_s: float = 60,
        tick_interval_s: float = .5,
        tick_fn: ta.Callable[..., bool] = lambda: True,
) -> ta.Mapping[ta.Callable, cf.Future]:
    for fn, deps in dependency_sets_by_fn.items():
        for dep in deps:
            if dep == fn:
                raise ValueError(fn)
            if dep not in dependency_sets_by_fn:
                raise KeyError(dep)
            if fn in dependency_sets_by_fn[dep]:
                raise Exception(f'Cyclic dependencies: {fn} <-> {dep}', fn, dep)

    dependent_sets_by_fn: dict[ta.Callable, set[ta.Callable]] = {fn: set() for fn in dependency_sets_by_fn}
    for fn, deps in dependency_sets_by_fn.items():
        for dep in deps:
            dependent_sets_by_fn[dep].add(fn)
    remaining_dep_sets_by_fn = {
        fn: set(dependencies) for fn, dependencies in dependency_sets_by_fn.items()
    }
    root_fns = {fn for fn, deps in remaining_dep_sets_by_fn.items() if not deps}
    fns_by_fut = {fut: fn for fn in root_fns for fut in [executor.submit(fn)] if fut is not None}

    def cancel():
        for cancel_fut in fns_by_fut:
            cancel_fut.cancel()

    start = time.time()
    not_done = set(fns_by_fut.keys())
    while not_done and tick_fn():
        done, not_done = cf.wait(not_done, timeout=tick_interval_s, return_when=cf.FIRST_COMPLETED)
        not_done = set(not_done)

        for fut in done:
            if fut.exception():
                cancel()
                raise FutureError(fut) from fut.exception()

            fn = fns_by_fut[fut]
            for dependent_fn in dependent_sets_by_fn.get(fn, set()):
                remaining_deps = remaining_dep_sets_by_fn[dependent_fn]
                remaining_deps.remove(fn)
                if not remaining_deps:
                    downstream_fut = executor.submit(dependent_fn)
                    if downstream_fut is not None:
                        fns_by_fut[downstream_fut] = dependent_fn
                        not_done.add(downstream_fut)

        if time.time() >= (start + timeout_s):
            cancel()
            raise FutureTimeoutError

    remaining_fns = {fn: deps for fn, deps in remaining_dep_sets_by_fn.items() if deps}
    if remaining_fns:
        raise Exception(f'Unfinished fns: {remaining_fns}', remaining_fns)

    futs_by_fn = {fn: fut for fut, fn in fns_by_fut.items()}
    return futs_by_fn



"""
TODO:
 - options lol - csv header, newline, etc
"""
import dataclasses as dc
import enum
import json
import typing as ta

from .... import lang


if ta.TYPE_CHECKING:
    import ast
    import csv
    import tomllib

    import yaml

    from ... import dotenv
    from ... import props
    from ... import xml

else:
    ast = lang.proxy_import('ast')
    csv = lang.proxy_import('csv')
    tomllib = lang.proxy_import('tomllib')

    yaml = lang.proxy_import('yaml')

    dotenv = lang.proxy_import('...dotenv', __package__)
    props = lang.proxy_import('...props', __package__)
    xml = lang.proxy_import('...xml', __package__)


##


@dc.dataclass(frozen=True)
class Format:
    names: ta.Sequence[str]
    load: ta.Callable[[ta.TextIO], ta.Any]


class Formats(enum.Enum):
    JSON = Format(['json'], json.load)

    YAML = Format(['yaml', 'yml'], lambda f: yaml.safe_load(f))

    TOML = Format(['toml'], lambda f: tomllib.loads(f.read()))

    ENV = Format(['env', 'dotenv'], lambda f: dotenv.dotenv_values(stream=f))

    PROPS = Format(['properties', 'props'], lambda f: dict(props.Properties().load(f.read())))

    PY = Format(['py', 'python', 'repr'], lambda f: ast.literal_eval(f.read()))

    XML = Format(['xml'], lambda f: xml.build_simple_element(xml.parse_tree(f.read()).getroot()).as_dict())

    CSV = Format(['csv'], lambda f: list(csv.DictReader(f)))
    TSV = Format(['tsv'], lambda f: list(csv.DictReader(f, delimiter='\t')))
    FLAT_CSV = Format(['fcsv'], lambda f: list(csv.reader(f)))
    FLAT_TSV = Format(['ftsv'], lambda f: list(csv.reader(f, delimiter='\t')))


FORMATS_BY_NAME: ta.Mapping[str, Format] = {
    n: f
    for e in Formats
    for f in [e.value]
    for n in f.names
}



import abc
import contextlib
import typing as ta

import pytest

from .... import lang
from ._registry import register


class ManagerMark(lang.Abstract):
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

    @abc.abstractmethod
    def __call__(self, item: pytest.Function) -> ta.Iterator[None]:
        raise NotImplementedError


def _deep_subclasses(cls):
    ret = set()

    def rec(cur):
        for nxt in cur.__subclasses__():
            if nxt not in ret:
                ret.add(nxt)
                rec(nxt)

    rec(cls)
    return ret


@register
class ManagerMarksPlugin:

    @lang.cached_function
    def mark_classes(self) -> ta.Mapping[str, type[ManagerMark]]:
        return {
            cls.__name__: cls
            for cls in lang.deep_subclasses(ManagerMark)  # type: ignore[type-abstract]
            if not lang.is_abstract_class(cls)
        }

    def pytest_configure(self, config):
        for n in self.mark_classes():
            config.addinivalue_line(
                'markers',
                f'{n}: mark to manage {n}',
            )

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_call(self, item):
        with contextlib.ExitStack() as es:
            for n, cls in self.mark_classes().items():
                if (m := item.get_closest_marker(n)) is None:
                    continue
                inst = cls(*m.args, **m.kwargs)
                es.enter_context(contextlib.contextmanager(inst)(item))

            yield
